function p=dmc(p)
%% DMC   Dynamic Matrix Control
% P=DMC(P) determines the dynamic matrix control (input change) based on
% the plant model (step response) and current measurement stored in the
% structure P.
% Input:
%   P.sr - unit step response data
%   P.u  - current input, initially 0
%   P.v  - past input, initially empty
%   P.G  - dynamic matrix, set by the initial call
%   P.F  - matrix to calculate free response, set by the initial call
%   P.k  - DMC gain, set by the initial call
%   P.r  - reference (set point)
%   P.a  - reference smooth factor
%   P.p  - prediction horizon
%   P.m  - moving horizon
%   P.y  - current mrasurement
%   P.la - performance criterion weight, i.e. J = ||r-y|| + p.la*||du||
%          where du is the input change
% Output:
%   P.u  - new input for next step
%   P.f  - updated free response
%   P.G  - dynamic matrix, if it is the first step.
%   P.k  - DMC gain, if it is the first step
%
% See Also: mpc

% Version 1.0 created by Yi Cao at Cranfield University on 6th April 2008.
% This version was modified by Saul Antolin at Universidad de Valladolid on
% 26 January 2024.

% Example:
%{
p.sr=filter([0 0 0.2713],[1 -0.8351],ones(50,1));
p.p=10;
p.m=5;
p.y=0;
p.v=[];
u=zeros(1,3);
N=120;
Y=zeros(N,1);
U=zeros(N,1);
R=zeros(N,1);
R([1:30 61:90])=1;
p.la=1;
for k=1:120
    p.a=0;
    p.r=R(k:min(N,k+p.p));
    if k>60
        p.a=0.7;
    end
    p=dmc(p);
    Y(k)=p.y;
    U(k)=p.u;
    u=[u(2:3) p.u];
    p.y=0.8351*p.y+0.2713*u(1);
end
subplot(211)
plot(1:N,Y,'b-',1:N,R,'r--',[60 60],[-0.5 1.5],':','linewidth',2)
title('solid: output, dashed: reference')
text(35,1,'\alpha=0')
text(95,1,'\alpha=0.7')
axis([0 120 -0.5 1.5])
subplot(212)
[xx,yy]=stairs(1:N,U);
plot(xx,yy,'-',[60 60],[-0.5 1.5],':','linewidth',2)
axis([0 120 -0.5 1.5])
title('input')
xlabel('time, min')
%}
% Input and output check
error(nargchk(1,1,nargin));
error(nargoutchk(0,1,nargout));

% Persistent var. for the change management of the prediction & control horizon
persistent past_prediction_horizont;
persistent past_control_horizont;
persistent past_criterion_weight;
persistent resp_salto_interna;

% length of step response
N=numel(p.sr);
P=p.p;

% initial setup
if isempty(p.v)
    % storage for past input
    p.v=zeros(N,1);
    % assing the step resp to the persistent var for operations
    resp_salto_interna=p.sr;
    % matrix to calculate free response from past input
    x=p.sr(1:N);
    resp_salto_interna(N+1:N+P)=resp_salto_interna(N);
    p.F=hankel(resp_salto_interna(2:P+1),resp_salto_interna(P+1:N+P))-repmat(x(:)',P,1);
    % dynamic matrix
    p.G=toeplitz(resp_salto_interna(1:P),resp_salto_interna(1)*eye(1,p.m));
    % calculate DMC gain
    R=chol(p.G'*p.G+p.la*eye(p.m));
    K=R\(R'\p.G');
    % only the first input will be used
    p.k=K(1,:);
    % Assign the value of MV only in the first iteration
    %p.u=0;
    p.u=p.iteracion_inicial;
    % Assign the value of pred. & control horizont to the persistent variable
    past_prediction_horizont=P;
    past_control_horizont=p.m;
    % Assign the value of performance criterion weight to the persistent variable
    past_criterion_weight=p.la;
end
if isempty(p.y)
    return
end

% Prediction horizont change management
if (past_prediction_horizont~=P)
    % assing the step resp to the persistent var for operations
    resp_salto_interna=p.sr;
    % matrix to calculate free response from past input
    x=p.sr(1:N);
    resp_salto_interna(N+1:N+P)=resp_salto_interna(N);
    p.F=hankel(resp_salto_interna(2:P+1),resp_salto_interna(P+1:N+P))-repmat(x(:)',P,1);
    % dynamic matrix
    p.G=toeplitz(resp_salto_interna(1:P),resp_salto_interna(1)*eye(1,p.m));
    % calculate DMC gain
    R=chol(p.G'*p.G+p.la*eye(p.m));
    K=R\(R'\p.G');
    % only the first input will be used
    p.k=K(1,:);
    % Assign the value of pred. horizont to the persistent variable
    past_prediction_horizont=P;
end

% Control horizont & criterion weight change management
if (past_control_horizont~=p.m || past_criterion_weight~=p.la)
    % dynamic matrix
    p.G=toeplitz(resp_salto_interna(1:P),resp_salto_interna(1)*eye(1,p.m));
    % calculate DMC gain
    R=chol(p.G'*p.G+p.la*eye(p.m));
    K=R\(R'\p.G');
    % only the first input will be used
    p.k=K(1,:);
    % Assign the value of control horizont to the persistent variable
    past_control_horizont=p.m;
    % Assign the value of performance criterion weight to the persistent variable
    past_criterion_weight=p.la;
end

% free response
f=p.F*p.v+p.y;
% smooth reference
nr=numel(p.r);
if nr>=P
    ref=p.r(1:P);
else
    ref=[p.r(:);p.r(end)+zeros(P-nr,1)];
end
w=filter([0 (1-p.a)],[1 -p.a],ref,p.y);
% DMC input change
u=p.k*(w-f);
% past input change for next step
p.v=[u;p.v(1:end-1)];
% next input
p.u=p.u+u(1);

end